open Client

(* -------------------------------------------------------------------------- *)

(* A wrapper over the client's [translate] function. We consider ill-typedness
   as normal, since our terms are randomly generated, so we translate the client
   exceptions to [None]. *)

let print_type ty =
  PPrint.(ToChannel.pretty 0.9 80 stdout (FPrinter.print_type ty ^^ hardline))

let translate e t =
  try
    Some (Infer.translate e t)
  with
  | Infer.Cycle (_range, ty) ->
      if Config.verbose then begin
        Printf.fprintf stdout "Type error: a cyclic type arose.\n";
        print_type ty
      end;
      None
  | Infer.Unify (_range, ty1, ty2) ->
      if Config.verbose then begin
        Printf.fprintf stdout "Type error: type mismatch.\n";
        Printf.fprintf stdout "Type error: mismatch between the type:\n";
        print_type ty1;
        Printf.fprintf stdout "and the type:\n";
        print_type ty2
      end;
      None

(* -------------------------------------------------------------------------- *)

(* Running all passes over a single ML term. *)

let test_with_args
  ~(success:int ref) ~(total:int ref) (env : ML.datatype_env) (t : ML.term)
  : bool
  =
  incr total;
  (* We print the term, parse it and compare with the given term *)
  let s = MLPrinter.to_string t in
  let lexbuf = Lexing.from_string s in
  let t' = MLParser.self_contained_term MLLexer.read lexbuf in
  assert (t = t');
  let outcome_env =
    ML2F.translate_env env in
  print_endline "Type inference and translation to System F...";
  let outcome =
    translate env t
  in
  match outcome with
  | None ->
      (* This term is ill-typed. This is considered a normal outcome, since
         our terms can be randomly generated. *)
      true
  | Some (t : F.nominal_term) ->
      CheckF.test outcome_env t;
      (* Everything seems to be OK. *)
      incr success;
      true

let test env t =
  test_with_args ~success:(ref 0) ~total:(ref 0) env t
