open Client

(* ∀a b. ({} -> (a * (a -> b))) -> b
   Λa b. fun (p : {} -> (a * (a -> b))) ->
     let (x, f) = p () in f x
*)
let let_prod =
  let open F in
  let alpha, beta = 0, 1 in
  let p, f, x = "p", "f", "x" in
  TyAbs (alpha, TyAbs (beta,
    Abs (p, TyArrow (TyProduct [],
                     TyProduct [TyVar alpha; TyArrow (TyVar alpha, TyVar beta)]),
      LetProd ([x; f], App (Var p, Tuple []),
        App (Var f, Var x)))))

(* Inj<∀α. α → α + ∀βα. α → β> 0 Λα.λx:α.x *)
let inj =
  let alpha, beta = 0, 1 in
  F.(
    Inj
      (
        [
          TyForall (alpha, TyArrow (TyVar alpha, TyVar alpha)) ;
          TyForall (beta, TyForall (alpha, TyArrow (TyVar alpha, TyVar beta)))
        ],
        0,
        TyAbs (alpha, Abs("x", TyVar alpha, Var "x"))
      )
  )

(* Λαβ. λ(x:{α*β}. match x with (_, y) -> y end *)
let match_with_prod =
  let alpha, beta = 1, 0 in
  F.(
    TyAbs(alpha,
      TyAbs(beta,
        Abs("x",
          TyProduct [ TyVar alpha ; TyVar beta ],
          Match(TyVar beta,
                Var "x",
                [
                  (PTuple [ PWildcard ; PVar "y" ] ,
                   Var "y")
                ]
            )
          )
        )
      )
  )

(* Λα. λ(x:α). match Inj<α+α> 0 x with Inj 0 y1 -> y1 | Inj 1 y2 -> y2 end *)
let match_with_sum =
  let alpha = 1 in
  F.(
    TyAbs(alpha,
      Abs("x",
        TyVar alpha,
        Match(TyVar alpha,
              Inj([TyVar alpha; TyVar alpha], 0, Var "x"),
              [
                (PInj([TyVar alpha; TyVar alpha], 0, PVar "y1"),
                 Var "y1");
                (PInj([TyVar alpha; TyVar alpha], 1, PVar "y2"),
                 Var "y2");
              ]
          )
        )
      )
  )

(* Λβ. match inj<0> () with
       | inj<0> () -> inj<0> ()
       | inj<1> (_) -> inj<0> ()
 *)
let match_inj = F.(
  let alpha = 25 in
  let beta = 308 in
  let option_type = [ TyProduct []; TyProduct [TyVar beta] ] in
  let none = Inj ( option_type, 0, Tuple []) in
    TyAbs(beta,
      Match (
        TyMu (alpha, TySum option_type ),
        none,
        [
          PInj ( option_type, 0, PTuple []),
          none
        ;
          PInj ( option_type, 1, PTuple [PWildcard]),
          none
        ]
      )
    )
  )

(* option *)
let option_env =
  let option_typedecl =
    let open Datatype in {
        name = Type "option";
        type_params = [ 0 ];
        data_kind = Variant;
        labels_descr = [
          {
            label_name = Label "None";
            type_name = Type "option";
            arg_type = F.TyProduct [];
          } ; {
            label_name = Label "Some";
            type_name = Type "option";
            arg_type = F.TyVar 0;
          }
        ];
      }
  in
  Datatype.Env.add_decl Datatype.Env.empty option_typedecl

let unit = F.Tuple []
let unit_pattern = F.PTuple []

(* None *)
let none =
  let alpha = 0 in
  F.(
    Variant (Datatype.Label "None", (Datatype.Type "option", [TyForall (alpha, TyVar alpha)]), unit)
  )

let none_pattern =
  let alpha = 0 in
  F.(
    PVariant (
      Datatype.Label "None",
      (Datatype.Type "option", [TyForall (alpha, TyVar alpha)]),
      unit_pattern
    )
  )

(* Some Λα.λx:α.x *)
let some =
  let alpha = 0 in
  F.(
    Variant (
      Datatype.Label "Some",
      (Datatype.Type "option", [ TyForall (alpha, TyArrow (TyVar alpha, TyVar alpha)) ]),
      TyAbs (alpha, Abs ("x", TyVar alpha, Var "x"))
    )
  )

let some_pattern =
  let alpha = 0 in
  F.(
    PVariant (
      Datatype.Label "Some",
      (Datatype.Type "option", [ TyForall (alpha, TyVar alpha) ]),
      PWildcard
    )
  )

(* Λα. match None with
       | None     -> ()
       | Some (_) -> ()
 *)
let match_none = F.(
  let alpha = 0 in
  TyAbs(alpha,
    Match (
      TyProduct [] ,
      none,
      [
        (none_pattern , unit)
      ;
        (some_pattern , unit)
      ]
      )
    )
  )

let test_ok_from_ast datatype_env t () =
  Alcotest.(check unit) "type inference" (Test.CheckF.test datatype_env t) ()

let test_case msg datatype_env t =
  Alcotest.(test_case msg `Quick (test_ok_from_ast datatype_env t))

let test_suite =
  [(
    "test F ast",
    [ test_case "let prod" Datatype.Env.empty let_prod;
      test_case "inj" Datatype.Env.empty inj;
      test_case "match with prod" Datatype.Env.empty match_with_prod;
      test_case "match with sum" Datatype.Env.empty match_with_sum;
      test_case "match inj" Datatype.Env.empty match_inj;
      test_case "unit" option_env unit;
      test_case "none" option_env none;
      test_case "some" option_env some;
      test_case "match none" option_env match_none;
    ]
  )]

let () =
  Alcotest.run "F test suite" test_suite
